close all;clear; clc;

USE_RTI = false;

% initial state
x0 = [0; pi; 0; 0];

%% OCP DESCRIPTION
ocp = AcadosOcp();

%% IVP DESCRIPTION
sim = AcadosSim();

%% SOLVER OPTIONS

%% discretization
T = .8; % OCP prediction horizon
N = 40; % number of shooting intervals
h = T / N; % discrete time step length

ocp.solver_options.tf = T;
ocp.solver_options.N_horizon = N;
if USE_RTI
    ocp.solver_options.nlp_solver_type = 'SQP_RTI';
else
    ocp.solver_options.nlp_solver_type = 'SQP';
end
ocp.solver_options.qp_solver = 'PARTIAL_CONDENSING_HPIPM';
% FULL_CONDENSING_HPIPM, PARTIAL_CONDENSING_HPIPM
% FULL_CONDENSING_QPOASES, PARTIAL_CONDENSING_OSQP
ocp.solver_options.qp_solver_cond_N = 5; % for partial condensing
% ocp.solver_options.globalization = 'MERIT_BACKTRACKING'; % turns on globalization

% integrator model
ocp.solver_options.integrator_type = 'IRK';

% integrator plant
sim.solver_options.Tsim = h;
sim.solver_options.integrator_type = 'IRK';

%% MODEL with mass as parameter
model = get_cartpole_model();
ocp.model = model;
sim.model = model;

% dimensions
nx = model.x.rows();
nu = model.u.rows();

%% COST: nonlinear-least squares cost
ocp.cost.cost_type_0 = 'NONLINEAR_LS';
ocp.cost.cost_type = 'NONLINEAR_LS';
ocp.cost.cost_type_e = 'NONLINEAR_LS';

W_x = diag([1e3, 1e3, 1e-2, 1e-2]);
W_u = 1e-2;

ocp.cost.yref_0 = zeros(nu, 1);
model.cost_y_expr_0 = model.u;
ocp.cost.yref = zeros(nx + nu, 1);
model.cost_y_expr = vertcat(model.x, model.u);
ocp.cost.yref_e = zeros(nx, 1);
model.cost_y_expr_e = model.x;

ocp.cost.W_0 = W_u;
ocp.cost.W = blkdiag(W_x, W_u);
ocp.cost.W_e = W_x;


%% CONSTRAINTS

U_max = 80;
ocp.constraints.idxbu = 0;
ocp.constraints.lbu = -U_max;
ocp.constraints.ubu = U_max;

ocp.constraints.x0 = x0;

%% OCP SOLVER
ocp_solver = AcadosOcpSolver(ocp);

%% SIM SOLVER/INTEGRATOR
sim_solver = AcadosSimSolver(sim);

%% SIMULATION
N_sim = 100;

% preallocate memory
x_sim = zeros(nx, N_sim+1);
u_sim = zeros(nu, N_sim);
if USE_RTI
    t_preparation = zeros(1, N_sim);
    t_feedback = zeros(1, N_sim);
else
    t = zeros(1, N_sim);
end

x_sim(:,1) = x0;
ocp_solver.set('constr_x0', x0);

% perform some iterations to initialize solver
if USE_RTI
    num_iter_initial = 5;
else
    num_iter_initial = 1;
end
for i=1:num_iter_initial
    ocp_solver.solve();
end

for i=1:N_sim
    % update initial state
    % TODO: initial state for OCP solved in this iteration
    x0 = 

    if USE_RTI
        % preparation phase
        ocp_solver.set('rti_phase', 1);
        ocp_solver.solve();
        t_preparation(i) = ocp_solver.get('time_tot');

        % feedback phase
        ocp_solver.set('rti_phase', 2);
        ocp_solver.set('constr_x0', x0);
        ocp_solver.solve();
        t_feedback(i) = ocp_solver.get('time_tot');

    else
        ocp_solver.set('constr_x0', x0);

        % solve
        ocp_solver.solve();
        t(i) = ocp_solver.get('time_tot');
    end

    % get solution
    u0 = ocp_solver.get('u', 0);
    status = ocp_solver.get('status'); % 0 - success
    if status ~= 0
        warning('acados returned status %d\n', status);
    end

    % set initial state for the simulation
    % TODO: Simulate forward one step by setting current state and control
    sim_solver.set('x', );
    sim_solver.set('u', );

    % simulate one step
    sim_status = sim_solver.solve();

    % get simulated state
    x_sim(:,i+1) = sim_solver.get('xn');


    u_sim(:,i) = u0;
end

if USE_RTI
    % scale to milliseconds
    t_feedback = t_feedback * 1e3;
    t_preparation = t_preparation * 1e3;
    fprintf('Computation time preparation phase in ms: min %.3f, median %.3f, max %.3f\n', min(t_preparation), median(t_preparation), max(t_preparation));
    fprintf('Computation time feedback phase in ms: min %.3f, median %.3f, max %.3f\n', min(t_feedback), median(t_feedback), max(t_feedback));

else
    % scale to milliseconds
    t = t * 1e3;
    fprintf('Computation time in ms: min %.3f, median %.3f, max %.3f\n', min(t), median(t), max(t));
end

%% plots
ts = linspace(0, N_sim*h, N_sim+1);
figure; hold on;
States = {'p', 'theta', 'v', 'dtheta'};

for i=1:length(States)
    subplot(length(States), 1, i);
    grid on; hold on;
    plot(ts, x_sim(i,:));
    ylabel(States{i});
    xlabel('t [s]')
end

figure
stairs(ts, [u_sim'; u_sim(end)])
ylabel('F [N]')
xlabel('t [s]')
grid on
